import 'package:flutter/widgets.dart';
import 'package:microfrontends/src/resolver.dart';

import 'effect.dart';
import 'state_container_provider.dart';
import 'mutation.dart';
import 'query.dart';
import 'state_container.dart';

@protected
extension Queries on BuildContext {
  @protected
  StateContainerProviderElement? getContainerElement() =>
      this.getElementForInheritedWidgetOfExactType<StateContainerProvider>()
          as StateContainerProviderElement?;

  StateContainer<T> getContainerOfType<T>() {
    try {
      final el = getContainerElement();
      final container = el?.containers[T];
      return container! as StateContainer<T>;
    } catch (err) {
      throw new Exception('No StateContainer<$T> found');
    }
  }

  Resolver getResolverForType<R, T>() {
    final el = getContainerElement();
    Resolver? r;

    if (R == QueryResolver) {
      r = el?.queryResolvers.getValueByTypeKey<T>(nullable: null is! T);
    } else {
      r = el?.mutationResolvers[T];
    }

    if (r == null) {
      throw new Exception('No resolver found for $T');
    }

    return r;
  }

  QueryResolver getResolverForQueryType<T>() {
    final r = getResolverForType<QueryResolver, T>();
    return r as QueryResolver;
  }

  MutationResolver getMutationResolverOfType<T>() {
    final r = getResolverForType<MutationResolver, T>();
    return r as MutationResolver;
  }

  StateContainerProviderElement getContainerHolder(container) {
    final el = getContainerElement();
    return el!.elements[container.runtimeType]!;
  }
}

extension FrameworkOps on BuildContext {
  T? query<T>([Object? params]) {
    return Query<T>(params).exec(this);
  }

  void mutation<T extends Mutation>(T mutation) {
    mutation.exec<T>(this);
  }

  void effect<T>(
    EffectCallback<T> callback, {
    leading = false,
  }) {
    final e = null is T
        ? Effect.nullable<T>(callback as EffectCallback<T?>)
        : Effect(callback);

    return e.bind(this, leading: leading);
  }
}

extension FrameworkWidgets on Widget {
  Widget withContainer<T>(
    BuildContext context, [
    StateContainer<T>? container,
  ]) {
    StateContainer<T> c = container ?? StateContainer.of<T>(context);

    return StateContainerProvider(
      container: c,
      child: this,
    );
  }
}
